#!/usr/bin/env python
# -*- coding: utf-8 -*-

# CAVEAT UTILITOR
#
# This file was automatically generated by Grako.
#
#    https://pypi.python.org/pypi/grako/
#
# Any changes you make to it will be overwritten the next time
# the file is generated.


from __future__ import print_function, division, absolute_import, unicode_literals
from grako.parsing import graken, Parser


__version__ = (2015, 1, 2, 21, 49, 31, 4)

__all__ = [
    'samtoolsParser',
    'samtoolsSemantics',
    'main'
]


class samtoolsParser(Parser):
    def __init__(self, whitespace='', nameguard=True, **kwargs):
        super(samtoolsParser, self).__init__(
            whitespace=whitespace,
            nameguard=nameguard,
            **kwargs
        )

    @graken()
    def _start_(self):
        with self._choice():
            with self._option():
                self._samtools_()
            with self._option():
                self._samtools_00_()
            with self._option():
                self._samtools_01_()
            with self._option():
                self._samtools_02_()
            with self._option():
                self._samtools_03_()
            self._error('no available options')

    @graken()
    def _samtools_(self):
        self._token('samtools')
        self._pattern(r'\s+')
        self._token('view')

        def block0():
            self._pattern(r'\s+')
            self._viewoptions_()
        self._closure(block0)
        self._pattern(r'\s+')
        self._inputfile_()

        def block1():
            self._pattern(r'\s+')
            self._regions_()
        self._closure(block1)
        self._check_eof()

    @graken()
    def _samtools_00_(self):
        self._token('samtools')
        self._pattern(r'\s+')
        self._token('flagstat')
        self._pattern(r'\s+')
        self._inputfile_()
        self._check_eof()

    @graken()
    def _samtools_01_(self):
        self._token('samtools')
        self._pattern(r'\s+')
        self._token('index')
        self._pattern(r'\s+')
        self._inputfile_()
        self._check_eof()

    @graken()
    def _samtools_02_(self):
        self._token('samtools')
        self._pattern(r'\s+')
        self._token('merge')
        self._pattern(r'\s+')
        self._outputfile_()
        self._pattern(r'\s+')
        self._inputfile_()

        def block0():
            self._pattern(r'\s+')
            self._inputfile_()
        self._closure(block0)
        self._check_eof()

    @graken()
    def _samtools_03_(self):
        self._token('samtools')
        self._pattern(r'\s+')
        self._token('sort')

        def block0():
            self._pattern(r'\s+')
            self._sortoptions_()
        self._closure(block0)
        self._pattern(r'\s+')
        self._inputfile_()
        self._pattern(r'\s+')
        self._outprefix_()
        self._check_eof()

    @graken()
    def _sortoptions_(self):
        with self._group():
            with self._choice():
                with self._option():
                    self._sortoptions_flag_()
                with self._option():
                    self._token('-l')
                    self._pattern(r'\s+')
                    self._int_()
                    self.ast.setlist('compression_level', self.last_node)
                with self._option():
                    self._token('-@')
                    self._pattern(r'\s+')
                    self._int_()
                    self.ast.setlist('no_threads', self.last_node)
                with self._option():
                    self._token('-m')
                    self._pattern(r'\s+')
                    self._string_()
                    self.ast.setlist('thread_memory', self.last_node)
                self._error('no available options')

        self.ast._define(
            [],
            ['compression_level', 'no_threads', 'thread_memory']
        )

    @graken()
    def _sortoptions_flag_(self):
        self._token('-')

        def block0():
            with self._choice():
                with self._option():
                    self._token('n')
                    self.ast.setlist('sort_by_read', self.last_node)
                with self._option():
                    self._token('f')
                    self.ast.setlist('prefix_is_fullname', self.last_node)
                with self._option():
                    self._token('o')
                    self.ast.setlist('output_to_stdout', self.last_node)
                self._error('expecting one of: f n o')
        self._positive_closure(block0)

        self.ast._define(
            [],
            ['sort_by_read', 'prefix_is_fullname', 'output_to_stdout']
        )

    @graken()
    def _viewoptions_(self):
        with self._group():
            with self._choice():
                with self._option():
                    self._viewoptions_flag_()
                with self._option():
                    self._token('-@')
                    self._pattern(r'\s+')
                    self._int_()
                with self._option():
                    self._token('-L')
                    self._pattern(r'\s+')
                    self._bedoverlap_()
                with self._option():
                    self._token('-t')
                    self._pattern(r'\s+')
                    self._file_()
                with self._option():
                    self._token('-T')
                    self._pattern(r'\s+')
                    self._file_()
                with self._option():
                    self._token('-o')
                    self._pattern(r'\s+')
                    self._file_()
                with self._option():
                    self._token('-R')
                    self._pattern(r'\s+')
                    self._file_()
                with self._option():
                    self._token('-f')
                    self._pattern(r'\s+')
                    self._int_()
                with self._option():
                    self._token('-F')
                    self._pattern(r'\s+')
                    self._int_()
                with self._option():
                    self._token('-q')
                    self._pattern(r'\s+')
                    self._int_()
                with self._option():
                    self._token('-l')
                    self._pattern(r'\s+')
                    self._string_()
                with self._option():
                    self._token('-r')
                    self._pattern(r'\s+')
                    self._string_()
                with self._option():
                    self._token('-s')
                    self._pattern(r'\s+')
                    self._float_()
                    self.ast.setlist('frac_subsample', self.last_node)
                self._error('no available options')

        self.ast._define(
            [],
            ['frac_subsample']
        )

    @graken()
    def _viewoptions_flag_(self):
        self._token('-')

        def block0():
            with self._choice():
                with self._option():
                    self._token('b')
                    self.ast.setlist('output_is_bam', self.last_node)
                with self._option():
                    self._token('h')
                    self.ast.setlist('output_header', self.last_node)
                with self._option():
                    self._token('H')
                with self._option():
                    self._token('S')
                    self.ast.setlist('input_is_sam', self.last_node)
                with self._option():
                    self._token('u')
                with self._option():
                    self._token('1')
                with self._option():
                    self._token('x')
                with self._option():
                    self._token('X')
                with self._option():
                    self._token('c')
                with self._option():
                    self._token('B')
                with self._option():
                    self._token('?')
                self._error('expecting one of: 1 ? B H S X b c h u x')
        self._positive_closure(block0)

        self.ast._define(
            [],
            ['output_is_bam', 'output_header', 'input_is_sam']
        )

    @graken()
    def _outputfile_(self):
        self._pattern(r'\S+\.bam')
        self.ast.setlist('outputfile', self.last_node)

        self.ast._define(
            [],
            ['outputfile']
        )

    @graken()
    def _inputfile_(self):
        self._pattern(r'\S+\.[bs]am')
        self.ast.setlist('inputfile', self.last_node)

        self.ast._define(
            [],
            ['inputfile']
        )

    @graken()
    def _bedoverlap_(self):
        self._file_()
        self.ast.setlist('bedoverlap', self.last_node)

        self.ast._define(
            [],
            ['bedoverlap']
        )

    @graken()
    def _regions_(self):
        self._pattern(r'\S+')
        self.ast.setlist('regions', self.last_node)

        self.ast._define(
            [],
            ['regions']
        )

    @graken()
    def _outprefix_(self):
        self._pattern(r'\S+')
        self.ast.setlist('outprefix', self.last_node)

        self.ast._define(
            [],
            ['outprefix']
        )

    @graken()
    def _sort_dasho_(self):
        self._token('-o')
        self.ast.setlist('sort_dasho', self.last_node)

        self.ast._define(
            [],
            ['sort_dasho']
        )

    @graken()
    def _file_(self):
        self._pattern(r'\S+')

    @graken()
    def _path_(self):
        self._file_()

    @graken()
    def _int_(self):
        self._pattern(r'[0-9]+')

    @graken()
    def _integer_(self):
        self._int_()

    @graken()
    def _float_(self):
        self._pattern(r'[0-9\.]+')

    @graken()
    def _string_(self):
        with self._group():
            with self._choice():
                with self._option():
                    self._pattern(r'"[^"]*"')
                with self._option():
                    self._pattern(r"'[^']*'")
                with self._option():
                    self._pattern(r'\S+')
                self._error('expecting one of: "[^"]*" \'[^\']*\' \\S+')

    @graken()
    def _str_(self):
        self._string_()


class samtoolsSemantics(object):
    def start(self, ast):
        return ast

    def samtools(self, ast):
        return ast

    def samtools_00(self, ast):
        return ast

    def samtools_01(self, ast):
        return ast

    def samtools_02(self, ast):
        return ast

    def samtools_03(self, ast):
        return ast

    def sortoptions(self, ast):
        return ast

    def sortoptions_flag(self, ast):
        return ast

    def viewoptions(self, ast):
        return ast

    def viewoptions_flag(self, ast):
        return ast

    def outputfile(self, ast):
        return ast

    def inputfile(self, ast):
        return ast

    def bedoverlap(self, ast):
        return ast

    def regions(self, ast):
        return ast

    def outprefix(self, ast):
        return ast

    def sort_dasho(self, ast):
        return ast

    def file(self, ast):
        return ast

    def path(self, ast):
        return ast

    def int(self, ast):
        return ast

    def integer(self, ast):
        return ast

    def float(self, ast):
        return ast

    def string(self, ast):
        return ast

    def str(self, ast):
        return ast


def main(filename, startrule, trace=False, whitespace=None, nameguard=None):
    import json
    with open(filename) as f:
        text = f.read()
    parser = samtoolsParser(parseinfo=False)
    ast = parser.parse(
        text,
        startrule,
        filename=filename,
        trace=trace,
        whitespace=whitespace,
        nameguard=nameguard)
    print('AST:')
    print(ast)
    print()
    print('JSON:')
    print(json.dumps(ast, indent=2))
    print()

if __name__ == '__main__':
    import argparse
    import string
    import sys

    class ListRules(argparse.Action):
        def __call__(self, parser, namespace, values, option_string):
            print('Rules:')
            for r in samtoolsParser.rule_list():
                print(r)
            print()
            sys.exit(0)

    parser = argparse.ArgumentParser(description="Simple parser for samtools.")
    parser.add_argument('-l', '--list', action=ListRules, nargs=0,
                        help="list all rules and exit")
    parser.add_argument('-n', '--no-nameguard', action='store_true',
                        dest='no_nameguard',
                        help="disable the 'nameguard' feature")
    parser.add_argument('-t', '--trace', action='store_true',
                        help="output trace information")
    parser.add_argument('-w', '--whitespace', type=str, default=string.whitespace,
                        help="whitespace specification")
    parser.add_argument('file', metavar="FILE", help="the input file to parse")
    parser.add_argument('startrule', metavar="STARTRULE",
                        help="the start rule for parsing")
    args = parser.parse_args()

    main(
        args.file,
        args.startrule,
        trace=args.trace,
        whitespace=args.whitespace,
        nameguard=not args.no_nameguard
    )




#!/usr/bin/env python
# -*- coding: utf-8 -*-

# CAVEAT UTILITOR
#
# This file was automatically generated by Grako.
#
#    https://pypi.python.org/pypi/grako/
#
# Any changes you make to it will be overwritten the next time
# the file is generated.


from __future__ import print_function, division, absolute_import, unicode_literals
from grako.parsing import graken, Parser


__version__ = (2015, 1, 2, 21, 49, 30, 4)

__all__ = [
    'bowtie2Parser',
    'bowtie2Semantics',
    'main'
]


class bowtie2Parser(Parser):
    def __init__(self, whitespace='', nameguard=True, **kwargs):
        super(bowtie2Parser, self).__init__(
            whitespace=whitespace,
            nameguard=nameguard,
            **kwargs
        )

    @graken()
    def _start_(self):
        with self._choice():
            with self._option():
                self._bowtie2_()
            with self._option():
                self._bowtie2_00_()
            with self._option():
                self._bowtie2_01_()
            with self._option():
                self._bowtie2_02_()
            with self._option():
                self._bowtie2_03_()
            with self._option():
                self._bowtie2_04_()
            self._error('no available options')

    @graken()
    def _bowtie2_(self):
        self._token('bowtie2')
        self._check_eof()

    @graken()
    def _bowtie2_00_(self):
        self._token('bowtie2')

        def block0():
            self._pattern(r'\s+')
            self._bowtieoptions_()
        self._closure(block0)
        self._pattern(r'\s+')
        self._token('-x')
        self._pattern(r'\s+')
        self._database_()
        self._pattern(r'\s+')
        self._token('-1')
        self._pattern(r'\s+')
        self._forward_reads_()
        self._pattern(r'\s+')
        self._token('-2')
        self._pattern(r'\s+')
        self._reverse_reads_()
        self._pattern(r'\s+')
        self._token('-S')
        self._pattern(r'\s+')
        self._samout_()
        self._check_eof()

    @graken()
    def _bowtie2_01_(self):
        self._token('bowtie2')

        def block0():
            self._pattern(r'\s+')
            self._bowtieoptions_()
        self._closure(block0)
        self._pattern(r'\s+')
        self._token('-x')
        self._pattern(r'\s+')
        self._database_()
        self._pattern(r'\s+')
        self._token('-1')
        self._pattern(r'\s+')
        self._forward_reads_()
        self._pattern(r'\s+')
        self._token('-2')
        self._pattern(r'\s+')
        self._reverse_reads_()
        self._check_eof()

    @graken()
    def _bowtie2_02_(self):
        self._token('bowtie2')

        def block0():
            self._pattern(r'\s+')
            self._bowtieoptions_()
        self._closure(block0)
        self._pattern(r'\s+')
        self._token('-x')
        self._pattern(r'\s+')
        self._database_()
        self._pattern(r'\s+')
        self._token('-U')
        self._pattern(r'\s+')
        self._reads_()
        self._pattern(r'\s+')
        self._token('-S')
        self._pattern(r'\s+')
        self._samout_()
        self._check_eof()

    @graken()
    def _bowtie2_03_(self):
        self._token('bowtie2')

        def block0():
            self._pattern(r'\s+')
            self._bowtieoptions_()
        self._closure(block0)
        self._pattern(r'\s+')
        self._token('-x')
        self._pattern(r'\s+')
        self._database_()
        self._pattern(r'\s+')
        self._token('-U')
        self._pattern(r'\s+')
        self._reads_()
        self._check_eof()

    @graken()
    def _bowtie2_04_(self):
        self._token('bowtie2')
        self._check_eof()

    @graken()
    def _samout_(self):
        self._pattern(r'\S+')
        self.ast.setlist('samout', self.last_node)

        self.ast._define(
            [],
            ['samout']
        )

    @graken()
    def _forward_reads_(self):
        self._pattern(r'\S+')
        self.ast.setlist('forward_reads', self.last_node)

        self.ast._define(
            [],
            ['forward_reads']
        )

    @graken()
    def _reverse_reads_(self):
        self._pattern(r'\S+')
        self.ast.setlist('reverse_reads', self.last_node)

        self.ast._define(
            [],
            ['reverse_reads']
        )

    @graken()
    def _reads_(self):
        self._pattern(r'\S+')
        self.ast.setlist('reads', self.last_node)

        self.ast._define(
            [],
            ['reads']
        )

    @graken()
    def _database_(self):
        self._pattern(r'\S+')
        self.ast.setlist('database', self.last_node)

        self.ast._define(
            [],
            ['database']
        )

    @graken()
    def _bowtieoptions_(self):
        with self._group():
            with self._choice():
                with self._option():
                    self._bowtieoptions_flag_()
                with self._option():
                    self._token('--qseq')
                with self._option():
                    self._token('-s')
                    self._pattern(r'\s+')
                    self._int_()
                with self._option():
                    self._token('--skip')
                    self._pattern(r'\s+')
                    self._int_()
                with self._option():
                    self._token('-u')
                    self._pattern(r'\s+')
                    self._int_()
                with self._option():
                    self._token('--upto')
                    self._pattern(r'\s+')
                    self._int_()
                with self._option():
                    self._token('-5')
                    self._pattern(r'\s+')
                    self._int_()
                with self._option():
                    self._token('--trim5')
                    self._pattern(r'\s+')
                    self._int_()
                with self._option():
                    self._token('-3')
                    self._pattern(r'\s+')
                    self._int_()
                with self._option():
                    self._token('--trim3')
                    self._pattern(r'\s+')
                    self._int_()
                with self._option():
                    self._token('--phred33')
                with self._option():
                    self._token('--phred64')
                with self._option():
                    self._token('--int-quals')
                with self._option():
                    self._token('--very-fast')
                with self._option():
                    self._token('--fast')
                with self._option():
                    self._token('--sensitive')
                with self._option():
                    self._token('--very-sensitive')
                with self._option():
                    self._token('--very-fast-local')
                with self._option():
                    self._token('--fast-local')
                with self._option():
                    self._token('--sensitive-local')
                with self._option():
                    self._token('--very-sensitive-local')
                with self._option():
                    self._token('-N')
                    self._pattern(r'\s+')
                    self._int_()
                with self._option():
                    self._token('-L')
                    self._pattern(r'\s+')
                    self._int_()
                with self._option():
                    self._token('-i')
                    self._pattern(r'\s+')
                    self._str_()
                with self._option():
                    self._token('--n-ceil')
                    self._pattern(r'\s+')
                    self._str_()
                with self._option():
                    self._token('--dpad')
                    self._pattern(r'\s+')
                    self._int_()
                with self._option():
                    self._token('--gbar')
                    self._pattern(r'\s+')
                    self._int_()
                with self._option():
                    self._token('--ignore-quals')
                with self._option():
                    self._token('--nofw')
                with self._option():
                    self._token('--norc')
                with self._option():
                    self._token('--no-1mm-upfront')
                with self._option():
                    self._token('--end-to-end')
                with self._option():
                    self._token('--local')
                with self._option():
                    self._token('--ma')
                    self._pattern(r'\s+')
                    self._int_()
                with self._option():
                    self._token('--mp')
                    self._pattern(r'\s+')
                    self._int_()
                with self._option():
                    self._token('--np')
                    self._pattern(r'\s+')
                    self._int_()
                with self._option():
                    self._token('--rdg')
                    self._pattern(r'\s+')
                    self._twoints_()
                with self._option():
                    self._token('--rfg')
                    self._pattern(r'\s+')
                    self._twoints_()
                with self._option():
                    self._token('--score-min')
                    self._pattern(r'\s+')
                    self._str_()
                with self._option():
                    self._token('-k')
                    self._pattern(r'\s+')
                    self._int_()
                with self._option():
                    self._token('--all')
                with self._option():
                    self._token('-D')
                    self._pattern(r'\s+')
                    self._int_()
                with self._option():
                    self._token('-R')
                    self._pattern(r'\s+')
                    self._int_()
                with self._option():
                    self._token('-I')
                    self._pattern(r'\s+')
                    self._int_()
                with self._option():
                    self._token('--minins')
                    self._pattern(r'\s+')
                    self._int_()
                with self._option():
                    self._token('-X')
                    self._pattern(r'\s+')
                    self._int_()
                with self._option():
                    self._token('--maxins')
                    self._pattern(r'\s+')
                    self._int_()
                with self._option():
                    self._token('--fr')
                with self._option():
                    self._token('--rf')
                with self._option():
                    self._token('--ff')
                with self._option():
                    self._token('--no-mixed')
                with self._option():
                    self._token('--no-discordant')
                with self._option():
                    self._token('--no-dovetail')
                with self._option():
                    self._token('--no-contain')
                with self._option():
                    self._token('--no-overlap')
                with self._option():
                    self._token('--time')
                with self._option():
                    self._token('--un')
                    self._pattern(r'\s+')
                    self._path_()
                with self._option():
                    self._token('--al')
                    self._pattern(r'\s+')
                    self._path_()
                with self._option():
                    self._token('--un-conc')
                    self._pattern(r'\s+')
                    self._path_()
                with self._option():
                    self._token('--al-conc')
                    self._pattern(r'\s+')
                    self._path_()
                with self._option():
                    self._token('--un-gz')
                    self._pattern(r'\s+')
                    self._path_()
                with self._option():
                    self._token('--quiet')
                with self._option():
                    self._token('--met-file')
                    self._pattern(r'\s+')
                    self._file_()
                with self._option():
                    self._token('--met-stderr')
                with self._option():
                    self._token('--met')
                    self._pattern(r'\s+')
                    self._int_()
                with self._option():
                    self._token('--no-unal')
                with self._option():
                    self._token('--no-head')
                with self._option():
                    self._token('--no-sq')
                with self._option():
                    self._token('--rg-id')
                    self._pattern(r'\s+')
                    self._str_()
                with self._option():
                    self._token('--rg')
                    self._pattern(r'\s+')
                    self._str_()
                with self._option():
                    self._token('--omit-sec-seq')
                with self._option():
                    self._token('-p')
                    self._pattern(r'\s+')
                    self._int_()
                with self._option():
                    self._token('--threads')
                    self._pattern(r'\s+')
                    self._int_()
                with self._option():
                    self._token('--reorder')
                with self._option():
                    self._token('--mm')
                with self._option():
                    self._token('--qc-filter')
                with self._option():
                    self._token('--seed')
                    self._pattern(r'\s+')
                    self._int_()
                with self._option():
                    self._token('--non-deterministic')
                with self._option():
                    self._token('--version')
                with self._option():
                    self._token('--help')
                self._error('expecting one of: --all --end-to-end --fast --fast-local --ff --fr --help --ignore-quals --int-quals --local --met-stderr --mm --no-1mm-upfront --no-contain --no-discordant --no-dovetail --no-head --no-mixed --no-overlap --no-sq --no-unal --nofw --non-deterministic --norc --omit-sec-seq --phred33 --phred64 --qc-filter --qseq --quiet --reorder --rf --sensitive --sensitive-local --time --version --very-fast --very-fast-local --very-sensitive --very-sensitive-local')

    @graken()
    def _bowtieoptions_flag_(self):
        self._token('-')

        def block0():
            with self._choice():
                with self._option():
                    self._token('q')
                with self._option():
                    self._token('f')
                with self._option():
                    self._token('r')
                with self._option():
                    self._token('c')
                with self._option():
                    self._token('a')
                with self._option():
                    self._token('t')
                with self._option():
                    self._token('h')
                self._error('expecting one of: a c f h q r t')
        self._positive_closure(block0)

    @graken()
    def _twoints_(self):
        self._pattern(r'[0-9]+,[0-9]')
        self.ast.setlist('twoints', self.last_node)

        self.ast._define(
            [],
            ['twoints']
        )

    @graken()
    def _file_(self):
        self._pattern(r'\S+')

    @graken()
    def _path_(self):
        self._file_()

    @graken()
    def _int_(self):
        self._pattern(r'[0-9]+')

    @graken()
    def _integer_(self):
        self._int_()

    @graken()
    def _float_(self):
        self._pattern(r'[0-9\.]+')

    @graken()
    def _string_(self):
        with self._group():
            with self._choice():
                with self._option():
                    self._pattern(r'"[^"]*"')
                with self._option():
                    self._pattern(r"'[^']*'")
                with self._option():
                    self._pattern(r'\S+')
                self._error('expecting one of: "[^"]*" \'[^\']*\' \\S+')

    @graken()
    def _str_(self):
        self._string_()


class bowtie2Semantics(object):
    def start(self, ast):
        return ast

    def bowtie2(self, ast):
        return ast

    def bowtie2_00(self, ast):
        return ast

    def bowtie2_01(self, ast):
        return ast

    def bowtie2_02(self, ast):
        return ast

    def bowtie2_03(self, ast):
        return ast

    def bowtie2_04(self, ast):
        return ast

    def samout(self, ast):
        return ast

    def forward_reads(self, ast):
        return ast

    def reverse_reads(self, ast):
        return ast

    def reads(self, ast):
        return ast

    def database(self, ast):
        return ast

    def bowtieoptions(self, ast):
        return ast

    def bowtieoptions_flag(self, ast):
        return ast

    def twoints(self, ast):
        return ast

    def file(self, ast):
        return ast

    def path(self, ast):
        return ast

    def int(self, ast):
        return ast

    def integer(self, ast):
        return ast

    def float(self, ast):
        return ast

    def string(self, ast):
        return ast

    def str(self, ast):
        return ast


def main(filename, startrule, trace=False, whitespace=None, nameguard=None):
    import json
    with open(filename) as f:
        text = f.read()
    parser = bowtie2Parser(parseinfo=False)
    ast = parser.parse(
        text,
        startrule,
        filename=filename,
        trace=trace,
        whitespace=whitespace,
        nameguard=nameguard)
    print('AST:')
    print(ast)
    print()
    print('JSON:')
    print(json.dumps(ast, indent=2))
    print()

if __name__ == '__main__':
    import argparse
    import string
    import sys

    class ListRules(argparse.Action):
        def __call__(self, parser, namespace, values, option_string):
            print('Rules:')
            for r in bowtie2Parser.rule_list():
                print(r)
            print()
            sys.exit(0)

    parser = argparse.ArgumentParser(description="Simple parser for bowtie2.")
    parser.add_argument('-l', '--list', action=ListRules, nargs=0,
                        help="list all rules and exit")
    parser.add_argument('-n', '--no-nameguard', action='store_true',
                        dest='no_nameguard',
                        help="disable the 'nameguard' feature")
    parser.add_argument('-t', '--trace', action='store_true',
                        help="output trace information")
    parser.add_argument('-w', '--whitespace', type=str, default=string.whitespace,
                        help="whitespace specification")
    parser.add_argument('file', metavar="FILE", help="the input file to parse")
    parser.add_argument('startrule', metavar="STARTRULE",
                        help="the start rule for parsing")
    args = parser.parse_args()

    main(
        args.file,
        args.startrule,
        trace=args.trace,
        whitespace=args.whitespace,
        nameguard=not args.no_nameguard
    )




meta = {'regions': {'pattern': '/\\S+/', 'type': 'other'}, 'frac_subsample': {'pattern': '-s <float>', 'type': 'option', 'name': 'frac_subsample'}, 'no_threads': {'pattern': '-@ <int>', 'type': 'option', 'name': 'no_threads'}, 'bedoverlap': {'category': 'input', 'pattern': 'file', 'type': 'file'}, 'output_is_bam': {'pattern': '-b', 'type': 'flag', 'name': 'output_is_bam'}, 'input_is_sam': {'pattern': '-S', 'type': 'flag', 'name': 'input_is_sam'}, 'output_header': {'pattern': '-h', 'type': 'flag', 'name': 'output_header'}, 'sort_dasho': {'pattern': '"-o"', 'type': 'other'}, 'prefix_is_fullname': {'pattern': '-f', 'type': 'flag', 'name': 'prefix_is_fullname'}, 'outputfile': {'category': 'output', 'default': '__stdout__', 'cardinality': '1', 'type': 'file', 'pattern': '/\\S+\\.bam/'}, 'compression_level': {'pattern': '-l <int>', 'type': 'option', 'name': 'compression_level'}, 'outprefix': {'pattern': '/\\S+/', 'cardinality': '1', 'type': 'other'}, 'sort_by_read': {'pattern': '-n', 'type': 'flag', 'name': 'sort_by_read'}, 'inputfile': {'category': 'input', 'pattern': '/\\S+\\.[bs]am/', 'type': 'file'}, 'output_to_stdout': {'pattern': '-o', 'type': 'flag', 'name': 'output_to_stdout'}, 'thread_memory': {'pattern': '-m <string>', 'type': 'option', 'name': 'thread_memory'}}


# tools to help register & identify commands

import logging
import os
import pprint
import subprocess as sp
import sys

from jinja2 import Template
import yaml

import leip
from leip import set_local_config

from kea.utils import find_executable, register_executable
from kea.utils import get_tool_conf, create_kea_link


lg = logging.getLogger(__name__)


@leip.arg('-V', '--version', help='version number')
@leip.arg('name', help='executable name to list')
@leip.command
def show(app, args):
    lg.info("show conf: %s", args.name)

    conf = get_tool_conf(app, args.name, args.version)

    del conf['versions']

    yaml.dump(conf, sys.stdout, default_flow_style=False)


@leip.subparser
def register(app, args):
    """
    register tools with kea
    """
    pass


@leip.arg('version', help='version to set default')
@leip.arg('name', help='executable name to list')
@leip.subcommand(register, "set_default")
def register_set_default(app, args):
    """
    List tool versions
    """
    tkey = args.version
    tdata = app.conf['app.{}.version.{}'.format(args.name, args.version)]
    print "set default:"
    print '{}: {}'.format(tkey, tdata['executable'])

    tv = tdata['version']
    if len(tv) < 80:
        print '  version: {}'.format(tv)
    else:
        i = 0
        while tv:
            if i == 0:
                print '  version: {}'.format(tv[:80])
                tv = tv[80:]
            else:
                print '     {}'.format(tv[:86])
                tv = tv[86:]
            i += 1

    set_local_config(app, 'app.{}.default_version'.format(args.name), tkey)


def print_tool_versions(app, appname):
    tool_version_data = app.conf['app.{}.versions'.format(appname)]
    def_version = app.conf['app.{}.default_version'.format(appname)]
    for tkey in tool_version_data:
        tdata = app.conf['app.{}.versions.{}'.format(appname, tkey)]
        print '{}: {}'.format(tkey, tdata['executable'])
        tv = tdata['version']
        if len(tv) < 80:
            print '  version: {}'.format(tdata['version'])
        else:
            i = 0
            while tv:
                if i == 0:
                    print '  version: {}'.format(tv[:80])
                    tv = tv[80:]
                else:
                    print '     {}'.format(tv[:86])
                    tv = tv[86:]
                i += 1
        if tkey == def_version:
            print("  default")


@leip.arg('name', help='executable name to list', nargs='?')
@leip.subcommand(register, "list")
def register_list(app, args):
    """
    List tool versions
    """
    if not args.name is None:
        print_tool_versions(app, args.name)
        return
    for tool in app.conf['app']:
        if 'versions' in app.conf['app.{}'.format(tool)]:
            print(tool)


@leip.subcommand(register, "create_execs")
def register_execs(app, args):
    """
    Create Kea executables
    """
    for tool in app.conf['app']:
        if 'versions' in app.conf['app.{}'.format(tool)]:
            print(tool)
            create_kea_link(app, tool)


@leip.flag('-d', '--set_default', help='set this executable as default')
@leip.arg('-V', '--version', help='version number')
@leip.arg('name', help='executable name to register')
@leip.subcommand(register, "add")
def register_add(app, args):

    execname = args.name
    if '/' in execname:
        execname = execname.split('/')[-1]

    lg.debug("register %s", execname)

    execs = list(find_executable(args.name))
    no_execs = len(execs)

    # make sure there is a link in the kea bin path
    keabin = os.path.expanduser(app.conf['bin_path'])
    if not os.path.exists(keabin):
        os.makedirs(keabin)

    if no_execs == 1 and args.version:
        register_executable(app, execname, execs[0], args.version,
                            args.set_default)
        return

    if no_execs > 1 and args.version:
        lg.error("multiple executables found & have only one version")
        for e in execs:
            lg.error("  - %s", e)
        exit(-1)

    if no_execs > 1 and args.is_default:
        lg.error("multiple executables found & cannot set them all as default")
        for e in execs:
            lg.error("  - %s", e)
        exit(-1)

    toolconf = get_tool_conf(app, execname, version=None)

    if not 'version_command' in toolconf:
        lg.error("No information on how to retrieve tool version")
        lg.error("Please specify using:")
        lg.error("   kea conf set app.%s.version_command '<command>'",
                 execname)
        exit(-1)

    version_command = toolconf['version_command']
    lg.debug("version command: %s", version_command)
    vc_template = Template(version_command)

    lg.info("version command: %s", version_command)

    for ex in execs:
        cmd = vc_template.render(executable=ex)
        P = sp.Popen(cmd, shell=True, stdout=sp.PIPE)
        o, e = P.communicate()
        version = o.strip()
        register_executable(app, execname, ex, version,
                            args.set_default)


    #lg.warning("Could not register (yet?)")



"""
Parse command lines using grako generated parsers
"""
import collections
import importlib
import logging
import pkg_resources
import re
import shlex
import time

import grako
import leip
import yaml

lg = logging.getLogger(__name__)
# lg.setLevel(logging.DEBUG)

#thanks: http://tinyurl.com/nttpj9
def convert_to_str(data):
    if isinstance(data, basestring):
        return str(data)
    elif isinstance(data, collections.Mapping):
        return dict(map(convert_to_str, data.iteritems()))
    elif isinstance(data, collections.Iterable):
        return type(data)(map(convert_to_str, data))
    else:
        return data

def astparse(ast, data={}):
    if isinstance(ast, list):
        [astparse(x, data) for x in ast]
    elif isinstance(ast, dict):
        data.update(convert_to_str(ast))
    elif isinstance(ast, (str, unicode)):
        pass
    else:
        print('x' * 80, ast)
    return data


class KeaSemantics:

    def __init__(self, meta):
        self.meta = meta
        self.options = collections.defaultdict(list)
        self.files = collections.defaultdict(dict)

        for k, m in meta.iteritems():
            if 'default' in m:
                if m['type'] == 'flag':
                    self.set_value(k, True)
                else:
                    self.set_value(k, m['default'])

    def set_value(self, k, v):

        if isinstance(v, collections.Mapping):
            v = map(str, v)
        elif isinstance(v, unicode):
            v = str(v)

        kmeta = self.meta[k]
        ktype = kmeta['type']
        kcard = kmeta.get('cardinality', None)
        if kcard is None:
            kcard = '1' if ktype == 'flag' else '+'

        if kcard == '+' and not isinstance(v, list):
            v = [v]
        if kcard == '1' and isinstance(v, list):
            v = v[0]

        if ktype in ['option', 'other', 'flag']:
            if kcard == '1': 
                self.options[k] = v
            else:
                self.options[k].extend(v)

        if ktype == 'file':
            print k, v, kcard, kmeta
            self.files[k]['category'] = kmeta['category']
            if kcard == '1':
                self.files[k]['path'] = v
            else:
                self.files[k]['path'] = self.files[k].get('path', []) + v

        

    def _default(self, ast):
        if not isinstance(ast, grako.ast.AST):
            return

        for k, v in ast.iteritems():
            if not isinstance(v, list):
                continue
            if len(v) == 0:
                continue
            if not k in self.meta:
                continue

        
            self.set_value(k, v)

PARSERS = {}

@leip.hook('pre_fire')
def parse_commandline(app, jinf):

    global MODS

    start = time.time()
    
    if app.name in PARSERS:
        parser = PARSER[app.name]
    else:
        modpath = 'kea.plugin.cl_parser.parsers.{}'.format(app.name)
        metamodpath = 'kea.plugin.cl_parser.parsers.{}_meta'.format(app.name)

        mod = importlib.import_module(modpath)
        modm = importlib.import_module(metamodpath)

        parser = getattr(mod, '{}Parser'.format(app.name))()
        parser.meta = modm.meta
        PARSERS[app.name] = parser
        
    cl = " ".join(app.cl)

    semantics = KeaSemantics(modm.meta)

    try:
        ast = parser.parse(cl, rule_name="start", semantics=semantics)
    except Exception as e:
        lg.warning("Cannot cl parse")
        raise

    if not 'files' in jinf:
        jinf['files'] = {}
    jinf['files'].update(semantics.files)
    if not 'optioins' in jinf:
        jinf['options'] = {}
    jinf['options'].update(semantics.options)

    stop = time.time()
    lg.debug("clparse took %s seconds", stop-start)





import leip
import sys

MADAPP = None


def flag_find(lst, flg, app):
    if not flg or not flg in lst:
        return []

    p_last = 0
    p = lst.index(flg)

    rv = []
    while p != -1:
        value = lst[p+1]
        madfile = app.get_madfile(value)
        rv.append(madfile)
        p_last = p
        try:
            p = lst.index(flg, p_last+1)
        except ValueError:
            break
    return rv


def find_input_file(app, info):
    ff_conf = app.conf.get('filefind')

    if not ff_conf:
        return

    iff = ff_conf['input_file_flag']
    off = ff_conf['output_file_flag']

    if not 'input_files' in info:
        info['input_files'] = []
    if not 'output_files' in info:
        info['output_files'] = []

    info['input_files'].extend(flag_find(sys.argv, iff, app))
    info['output_files'].extend(flag_find(sys.argv, off, app))


# @leip.hook('pre_run')
# def hook_pre_run(app):
#     ff_conf = app.conf.get('filefind')
#     print ff_conf


@leip.hook('pre_fire')
def hook_find_input_file(app, info):
    return find_input_file(app, info)






from collections import OrderedDict
import logging

import leip
from lockfile import FileLock

import kea.utils

lg = logging.getLogger(__name__)

def to_str(s):
    return str(s)


@leip.hook('pre_argparse')
def logger_arg_define(app):
    logger_group = app.parser.add_argument_group('Report plugin')

    logger_group.add_argument('-S', '--report_screen', action='store_true')
    logger_group.add_argument('-Y', '--report_yaml', action='store_true')
    logger_group.add_argument('-L', '--report_file', action='store_true')

    
@leip.hook('post_fire')
def log_screen(app, jinf):
    
    if app.args.report_yaml:
        import yaml
        
        if 'psutil_process' in jinf:
            del jinf['psutil_process']

        fn = '{}.report.yaml'.format(jinf['run_uid'])
        with open(fn, 'w') as F:
            yaml.safe_dump(dict(jinf), F)
    
    if not app.args.report_screen:
        return
        
    def dictprint(d, firstpre="", nextpre=""):
        if len(d) == 0:
            return
        mxkyln = max([len(x) for x in d.keys()] + [5])
        fs = '{:<' + str(mxkyln) + '} : {}'
        fp = '{:<' + str(mxkyln) + '} > '
        i = 0
        for k in sorted(d.keys()):
            if k in ['args']:
                continue
                
            v = d[k]
            v = kea.utils.make_pretty_kv(k, v)
            
            if isinstance(v, str) and not v.strip():
                continue

            i = i + 1
            pre = firstpre if i == 1 else nextpre
            if not isinstance(v, dict):
                print pre + fs.format(k, v)
            else:                
                bfp = pre + fp.format(k)
                bnp = nextpre + fp.format(' ')
                dictprint(v, bfp, bnp)
                
    print '--KEA-REPORT' + '-' * 68
    dictprint(jinf)
    print '-' * 80
    
@leip.hook('post_run')
def log_cl(app):
    if not app.args.report_file:
        return

    all_jinf = app.all_jinf
    try:
        with FileLock('kea.log'):
            for i, info in enumerate(all_jinf):
                with open('kea.log', 'a') as F:
                    F.write("-" * 80 + "\n")
                    for i in info:
                        F.write("{}: ".format(i))
                        val = info[i]
                        if i == 'cl':
                            F.write(" ".join(val) + "\n")
                        elif isinstance(val, list):
                            F.write("\n")
                            for lv in val:
                                F.write(' - {}\n'.format(to_str(lv)))
                        else:
                            F.write(" {}\n".format(to_str(val)))

    except Exception as e:
        lg.warning("Cannot write to log file (%s)", str(e))




#!/usr/bin/env python

from setuptools import setup, find_packages

# one line description
with open('DESCRIPTION') as F:
    description = F.read().strip()

# version number
with open('VERSION') as F:
    version = F.read().strip()

entry_points = {
    'console_scripts': [
        'kea = kea.cli:dispatch',
    ]}

setup(name='kea',
      version=version,
      description=description,
      author='Mark Fiers',
      author_email='mark.fiers.42@gmail.com',
      entry_points=entry_points,
      include_package_data=True,
      url='https://encrypted.google.com/#q=Kea&safe=off',
      packages=find_packages(),
      install_requires=[
          'Leip',
          'lockfile',
          'fantail',
          'humanize',
          'jinja2',
          'psutil',
          'toMaKe'
      ],
      classifiers=[
          'Development Status :: 4 - Beta',
          'Environment :: Console',
          'Operating System :: OS Independent',
          'Programming Language :: Python',
          'Programming Language :: Python :: 2',
          'Programming Language :: Python :: 2.7',
      ]
      )




import os
import sys

import leip

from kea import Kea


def dispatch():
    """
    Run the MadMax app
    """
    app.run()

#check if there is a kea command

command = None
for i in range(1, len(sys.argv)):
    if not sys.argv[i].startswith('-'):
        command = sys.argv[i]
        break


if command in ['conf', 'snipset', 'jobset', 'js', 'run', 'jobrun', 'jr', 'list_executors']:
    app = leip.app('kea', partial_parse = True)
else:
    app = Kea()




# kea runner

import argparse
import copy
from collections import OrderedDict
import logging
import os
import shlex
import subprocess as sp
import sys

import leip
import fantail


from kea.utils import get_tool_conf
from kea.plugin.register import print_tool_versions
from kea.cl_generator import basic_command_line_generator

lg = logging.getLogger(__name__)
#lg.setLevel(logging.DEBUG)

conf = leip.get_config('kea')

#leiplog = logging.getLogger('leip')
#leiplog.setLevel(logging.DEBUG)


class Kea(leip.app):

    def __init__(self, *args, **kwargs):

        # Call Leip - but we do not want the argparser:
        super(Kea, self).__init__('kea', disable_commands=True)

        self.all_jinf = [] #store job reports
        
        # replace the config by a stack so we can backfill
        self.conf = fantail.Fanstack([self.conf, fantail.Fantail()])

        # hack - if -v/--verbose is set - use that early:
        if '-vv' in sys.argv:
            lg.setLevel(logging.DEBUG)
        elif '-v' in sys.argv:
            lg.setLevel(logging.INFO)
            
        # another hack - early discovery of the executor
        # print(sys.argv.find('-x')
        if not '-x' in sys.argv:
            self.executor = self.conf['default_executor']
        else:
            xindx = sys.argv.index('-x')
            if xindx < len(sys.argv):
                self.executor = sys.argv[xindx + 1]
            else:
                self.executor = self.conf['default_executor']

            
        lg.debug("start kea initialization")

        # different hooks!
        self.hook_order = [
            'pre_argparse',
            'argparse',
            'post_argparse',
            'prepare',
            'pre_run',
            'run',
            'post_run',
            'finish']

        self.parser = argparse.ArgumentParser(add_help=False)
        self.discover(globals())


@leip.hook('pre_argparse')
def main_arg_define(app):

    app.parser.add_argument('-h', '--help', action='store_true',
                            help='Show this help and exit')
    app.parser.add_argument('-v', '--verbose', action='count')
    app.parser.add_argument('-U', '--uid',
                            help='unique identifier for this run')
    app.parser.add_argument('-x', '--executor', help='executor to use',
                            default=app.conf['default_executor'])
    app.parser.add_argument('-V', '--version', help='tool version')
    app.parser.add_argument('-o', '--stdout', help='save stdout to')
    app.parser.add_argument('-e', '--stderr', help='save stderr to')
    app.parser.add_argument('-n', '--jobstorun', help='no jobs to start',
                            type=int)

    
@leip.hook('argparse')
def kea_argparse(app):
    """
    Separate Kea arguments from tool arguments & feed the kea arguments
    to argparse
    """


    app.parser.add_argument('command', nargs='?')
    app.parser.add_argument('arg', nargs=argparse.REMAINDER)

    app.args = app.parser.parse_args()

    if app.args.help:
        kea_app = leip.app('kea', partial_parse=True)
        print "# Command mode\n"
        app.parser.print_help()
        print "\n# Utility mode\n"
        kea_app.parser.print_help()
        exit(0)

    app.cl_args = app.args.arg

    if not app.args.command:
        app.parser.print_usage()
        kea_app = leip.app('kea', partial_parse=True)
        kea_app.parser.print_usage()
        sys.exit(-1)
        
    cl = [app.args.command]
    
    if app.cl_args:
        cl.extend(app.cl_args)
        
    #special case - probably used quotes on the command line
    if len(cl) == 1 and ' ' in cl[0]:
        cl = shlex.split(cl[0])

    app.cl = cl
    app.name = os.path.basename(app.cl[0])
    app.conf['executable'] = app.cl[0]

    conf = get_tool_conf(app, app.name, app.args.version)
    app.conf.stack[1] = conf

    if app.args.verbose == 1:
        logging.getLogger().setLevel(logging.INFO)
    if app.args.verbose > 1 :
        logging.getLogger().setLevel(logging.DEBUG)

    #load args into the config object
    app.defargs = fantail.Fantail()
    app.defargs.update(app.conf.get('default'))
    app.defargs.update(app.conf.get('app.default.{}'.format(app.name)))
    for k in app.args.__dict__:
        v = getattr(app.args, k)
        if v is None: continue
        app.defargs[k] = v
        
    app.executor = app.args.executor
    lg.debug("Loaded config: %s",  app.name)


    
@leip.hook('run')
def run_kea(app):
    
    lg.debug("Start Kea run")
    executor_name = app.executor

    lg.debug("loading executor %s", executor_name)
    executor = app.conf['executors.{}'.format(executor_name)](app)

    app.all_jinf = []
    for jinf in basic_command_line_generator(app):
        app.all_jinf.append(jinf)

        jinf['executable'] = app.conf['executable']
        jinf['executor'] = executor_name
        cl = jinf['cl']
        lg.debug("command line arguments: %s", " ".join(cl))

        jinf['args'] = " ".join(app.cl_args)
        jinf['cwd'] = os.getcwd()

        executor.fire(jinf)
        
    executor.finish()
